/**
 * abssqr.com Inc.
 * Copyright (c) 2017-${YEAR} All Rights Reserved.
 */
package com.abssqr.plat.biz.validate;

import com.abssqr.plat.common.facade.validation.Bind;
import com.abssqr.plat.common.model.exception.AbssqrValidationException;
import org.springframework.util.ReflectionUtils;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Field校验抽象类
 *
 * @author zhenxuan.luo
 * @version $Id: AFieldValidator.java, v 0.1 2015年5月19日 上午10:31:23 zhenxuan.luo Exp $
 */
public abstract class AFieldValidator<A extends Annotation> {
    //一个校验注解类可能对应多个校验的类型
    
    /** Object类型和原始类型的映射 */
    private static Map<Class<?>, Class<?>> classPrimitiveMap = new HashMap<Class<?>, Class<?>>();
    
    static {
        classPrimitiveMap.put(Boolean.class, Boolean.TYPE);
        classPrimitiveMap.put(Byte.class, Byte.TYPE);
        classPrimitiveMap.put(Character.class, Character.TYPE);
        classPrimitiveMap.put(Double.class, Double.TYPE);
        classPrimitiveMap.put(Float.class, Float.TYPE);
        classPrimitiveMap.put(Integer.class, Integer.TYPE);
        classPrimitiveMap.put(Long.class, Long.TYPE);
        classPrimitiveMap.put(Short.class, Short.TYPE);
    }
    
    /** 校验注解类,用来注解需要校验的对象 例如:NotBlank */
    private final Class<A> annotationClass;
    
    /** 和注解类搭配的可以被校验的类型 例如:Integer.Type String */
    private final Set<Class<?>> verifiableClassSet;
    
    /**
     * 构造函数
     * @param annotationClass 注解类,用来注解需要校验的对象
     */
    public AFieldValidator(Class<A> annotationClass) {
        //参数为null
        if (annotationClass == null) {
            throwValidateException("构造器参数异常", annotationClass);
        }
    
        //校验注解没有设置Bind
        Bind bind = getBind(annotationClass);
        if (bind == null) {
            throwValidateException("Bind注解没有设置", annotationClass);
        }
    
        //获取校验注解中Bind中表示的可校验类型
        Class<?>[] verifiableClassArray = bind.verifiableClass();
        if (verifiableClassArray == null || verifiableClassArray.length == 0) {
            throwValidateException("Bind的校验器列表为空", annotationClass);
        }
        this.annotationClass = annotationClass;
    
        //由于校验注解中只能填 对象的class,基本数据类型的class无法手填,所以这里需要扩展一下
        Set<Class<?>> set = new HashSet<Class<?>>();
        for (Class<?> c : verifiableClassArray) {
            Class<?> primitiveClass = getPrimitiveClassIfExisted(c);
            if (primitiveClass != null) {
                set.add(primitiveClass);
            }
            set.add(c);
        }
    
        this.verifiableClassSet = set;
    }
    
    /**
     * 校验对象(validTargetObj)的某个字段(field)
     * @param validateService 对象校验器
     * @param validTargetObj 校验的对象
     * @param field 校验对象(validTargetObj)的某个字段
     * @return null:没有错误;反之返回FieldError对象
     * @throws IllegalAccessException
     */
    @SuppressWarnings("unchecked")
    public FieldError validate(final ValidateService validateService, final Object validTargetObj,
                               final Field field, final Annotation annotationObj)
            throws IllegalAccessException {
        // 当前字段存在本校验注解对象
        if (annotationObj != null) {
            ReflectionUtils.makeAccessible(field);
            // 注解的字段类型是否合理
            Class<?> type = field.getType();
            if (!fieldTypeVerifiable(type)) {
                throw new AbssqrValidationException("注解标注的字段类型和注解注册的校验类型不一致" + field + ":" + type
                                                    + ":" + verifiableClassSet);
            }
            // 取出字段的值进行校验
            Object validTargetFieldObj = field.get(validTargetObj);
            FieldError fieldError = new FieldError(validTargetObj.getClass().getName(),
                    field.getName(), validTargetFieldObj);
            // 校验字段是否合法
            boolean isValid = isValid(validateService, (A) annotationObj, validTargetFieldObj,
                    fieldError);
            if (!isValid) {
                return fieldError;
            }
        }
        return null;
    }
    
    /**
     * 给子类实现的具体校验方法
     * @param validateService 对象校验器
     * @param annotationObj 校验注解对象
     * @param validTargetFieldObj 需要校验的字段对象
     * @param errorConext 每个字段错误的存储类
     * @return true:校验通过;反之返回false
     */
    protected abstract boolean isValid(ValidateService validateService, A annotationObj,
                                       Object validTargetFieldObj, FieldError errorConext);
    
    /**
     * throw {@link AbssqrValidationException}
     * @param errorDesc
     * @param annotationClass 注解类,用来注解需要校验的对象
     */
    private void throwValidateException(String errorDesc, Class<A> annotationClass) {
        StringBuilder sb = new StringBuilder();
        sb.append("构造").append(this).append("异常\tannotationClass:").append(annotationClass)
                .append(errorDesc);
        throw new AbssqrValidationException(sb.toString());
    }
    
    /**
     * 字段类型是否为注解可校验的类型
     * @param type
     * @return true:字段类型是可以被校验的类型:反之返回false
     */
    private boolean fieldTypeVerifiable(Class<?> type) {
        if (verifiableClassSet.contains(type)) {
            return true;
        }
        for (Class<?> verifiableClass : verifiableClassSet) {
            if (verifiableClass.isAssignableFrom(type)) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * 通过Object找到对应的原始类型
     * @param c
     * @return null:没找到对应的原始类型;反之返回找到的Class对象
     */
    private Class<?> getPrimitiveClassIfExisted(Class<?> c) {
        return classPrimitiveMap.get(c);
    }
    
    /**
     * 获取校验注解的{@link Bind}标注对象,这个Bind限制了标注的校验实现类,目的是用来防止错误的绑定关系
     * @param annotationClass 校验注解
     */
    private Bind getBind(Class<A> annotationClass) {
        Annotation[] annotations = annotationClass.getAnnotations();
        for (Annotation annotation : annotations) {
            if (annotation instanceof Bind) {
                return (Bind) annotation;
            }
        }
        return null;
    }
    
    /**
     * Getter method for property <tt>annotationClass</tt>.
     *
     * @return property value of annotationClass
     */
    public Class<A> getAnnotationClass() {
        return annotationClass;
    }
}
